/*
 * pixel.cc
 *
 * Description of this file:
 *    Pixel processing functions definition of the xrddec library
 *
 * --------------------------------------------------------------------------
 *
 *    xrddec - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#include "common.h"

#ifdef __cplusplus
extern "C" {
#endif

/**
 * ===========================================================================
 * local & global variables (const tables)
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * partition map table
 */
const uint8_t g_partition_map_tab[] = {
    //  4      8          12          16          20   24          28   32          36   40   44   48           52   56   60   64
    PART_4x4,  PART_4x8,  255,        PART_4x16,  255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 4
    PART_8x4,  PART_8x8,  255,        PART_8x16,  255, 255,        255, PART_8x32,  255, 255, 255, 255,        255, 255, 255, 255,          // 8
    255,       255,       255,        PART_12x16, 255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 12
    PART_16x4, PART_16x8, PART_16x12, PART_16x16, 255, 255,        255, PART_16x32, 255, 255, 255, 255,        255, 255, 255, PART_16x64,   // 16
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 20
    255,       255,       255,        255,        255, 255,        255, PART_24x32, 255, 255, 255, 255,        255, 255, 255, 255,          // 24
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 28
    255,       PART_32x8, 255,        PART_32x16, 255, PART_32x24, 255, PART_32x32, 255, 255, 255, 255,        255, 255, 255, PART_32x64,   // 32
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 36
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 40
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 44
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, PART_48x64,   // 48
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 52
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 56
    255,       255,       255,        255,        255, 255,        255, 255,        255, 255, 255, 255,        255, 255, 255, 255,          // 60
    255,       255,       255,        PART_64x16, 255, 255,        255, PART_64x32, 255, 255, 255, PART_64x48, 255, 255, 255, PART_64x64    // 64
};


#define PIXEL_ADD_PS_C(w, h) \
static void avsd_pixel_add_ps_##w##x##h(pel_t *a, intptr_t dstride, const pel_t *b0, const coeff_t* b1, intptr_t sstride0, intptr_t sstride1)\
{\
    int x, y;\
    for (y = 0; y < h; y++) {\
        for (x = 0; x < w; x++) {\
            a[x] = (pel_t)XRDDEC_CLIP1(b0[x] + b1[x]);\
        }\
        b0 += sstride0;\
        b1 += sstride1;\
        a  += dstride;\
    }\
}

#define BLOCKCOPY_PP_C(w, h) \
static void avsd_blockcopy_pp_##w##x##h(pel_t *a, intptr_t stridea, const pel_t *b, intptr_t strideb)\
{\
    int x, y;\
    for (y = 0; y < h; y++) {\
        for (x = 0; x < w; x++) {\
            a[x] = b[x];\
        }\
        a += stridea;\
        b += strideb;\
    }\
}

#define BLOCKCOPY_SS_C(w, h) \
static void avsd_blockcopy_ss_##w##x##h(coeff_t* a, intptr_t stridea, const coeff_t* b, intptr_t strideb)\
{\
    int x, y;\
    for (y = 0; y < h; y++) {\
        for (x = 0; x < w; x++) {\
            a[x] = b[x];\
        }\
        a += stridea;\
        b += strideb;\
    }\
}

#define BLOCK_OP_C(w, h) \
    PIXEL_ADD_PS_C(w, h); \
    BLOCKCOPY_PP_C(w, h); \
    BLOCKCOPY_SS_C(w, h);

BLOCK_OP_C(64, 64)  /* 64x64 */
BLOCK_OP_C(64, 32)
BLOCK_OP_C(32, 64)
BLOCK_OP_C(64, 16)
BLOCK_OP_C(64, 48)
BLOCK_OP_C(16, 64)
BLOCK_OP_C(48, 64)
BLOCK_OP_C(32, 32)  /* 32x32 */
BLOCK_OP_C(32, 16)
BLOCK_OP_C(16, 32)
BLOCK_OP_C(32,  8)
BLOCK_OP_C(32, 24)
BLOCK_OP_C( 8, 32)
BLOCK_OP_C(24, 32)
BLOCK_OP_C(16, 16)  /* 16x16 */
BLOCK_OP_C(16,  8)
BLOCK_OP_C( 8, 16)
BLOCK_OP_C(16,  4)
BLOCK_OP_C(16, 12)
BLOCK_OP_C( 4, 16)
BLOCK_OP_C(12, 16)
BLOCK_OP_C( 8,  8)  /* 8x8 */
BLOCK_OP_C( 8,  4)
BLOCK_OP_C( 4,  8)
BLOCK_OP_C( 4,  4)  /* 4x4 */


#define ALL_LUMA_PU(name1, name2, cpu) \
    pixf->name1[PART_64x64] = avsd_ ## name2 ## _64x64 ## cpu;  /* 64x64 */ \
    pixf->name1[PART_64x32] = avsd_ ## name2 ## _64x32 ## cpu;\
    pixf->name1[PART_32x64] = avsd_ ## name2 ## _32x64 ## cpu;\
    pixf->name1[PART_64x16] = avsd_ ## name2 ## _64x16 ## cpu;\
    pixf->name1[PART_64x48] = avsd_ ## name2 ## _64x48 ## cpu;\
    pixf->name1[PART_16x64] = avsd_ ## name2 ## _16x64 ## cpu;\
    pixf->name1[PART_48x64] = avsd_ ## name2 ## _48x64 ## cpu;\
    pixf->name1[PART_32x32] = avsd_ ## name2 ## _32x32 ## cpu;  /* 32x32 */ \
    pixf->name1[PART_32x16] = avsd_ ## name2 ## _32x16 ## cpu;\
    pixf->name1[PART_16x32] = avsd_ ## name2 ## _16x32 ## cpu;\
    pixf->name1[PART_32x8 ] = avsd_ ## name2 ## _32x8  ## cpu;\
    pixf->name1[PART_32x24] = avsd_ ## name2 ## _32x24 ## cpu;\
    pixf->name1[PART_8x32 ] = avsd_ ## name2 ## _8x32  ## cpu;\
    pixf->name1[PART_24x32] = avsd_ ## name2 ## _24x32 ## cpu;\
    pixf->name1[PART_16x16] = avsd_ ## name2 ## _16x16 ## cpu;  /* 16x16 */ \
    pixf->name1[PART_16x8 ] = avsd_ ## name2 ## _16x8  ## cpu;\
    pixf->name1[PART_8x16 ] = avsd_ ## name2 ## _8x16  ## cpu;\
    pixf->name1[PART_16x4 ] = avsd_ ## name2 ## _16x4  ## cpu;\
    pixf->name1[PART_16x12] = avsd_ ## name2 ## _16x12 ## cpu;\
    pixf->name1[PART_4x16 ] = avsd_ ## name2 ## _4x16  ## cpu;\
    pixf->name1[PART_12x16] = avsd_ ## name2 ## _12x16 ## cpu;\
    pixf->name1[PART_8x8  ] = avsd_ ## name2 ## _8x8   ## cpu;  /* 8x8 */ \
    pixf->name1[PART_8x4  ] = avsd_ ## name2 ## _8x4   ## cpu;\
    pixf->name1[PART_4x8  ] = avsd_ ## name2 ## _4x8   ## cpu;\
    pixf->name1[PART_4x4  ] = avsd_ ## name2 ## _4x4   ## cpu  /* 4x4 */

void xrddec_pixel_init(uint32_t cpuid, ao_funcs_t* pixf)
{
    ALL_LUMA_PU(add_ps, pixel_add_ps,  );
    ALL_LUMA_PU(copy_pp, blockcopy_pp, );
    ALL_LUMA_PU(copy_ss, blockcopy_ss, );

}

#ifdef __cplusplus
}
#endif
