/*
 * osdep.h
 *
 * Description of this file:
 *    platform-specific code functions definition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#ifndef __XRD_OSDEP_H__
#define __XRD_OSDEP_H__


/**
 * ===========================================================================
 * includes
 * ===========================================================================
 */

/* ---------------------------------------------------------------------------
 * disable warning C4996: functions or variables may be unsafe.
 */
#if defined(_MSC_VER)
#define WIN32_LEAN_AND_MEAN
#define _CRT_NONSTDC_NO_DEPRECATE
#define _CRT_SECURE_NO_DEPRECATE
#define _CRT_SECURE_NO_WARNINGS
#include <intrin.h>
#include <windows.h>
#endif

#define _LARGEFILE_SOURCE       1
#define _FILE_OFFSET_BITS       64
#if defined(__ICL) || defined(_MSC_VER)
#include "configw.h"
#else
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#if HAVE_STDINT_H
#include <stdint.h>
#else
#include <inttypes.h>
#endif

#if defined(__INTEL_COMPILER)
#include <mathimf.h>
#else
#include <math.h>
#endif

#if defined(_MSC_VER) || defined(__INTEL_COMPILER)
#include <float.h>
#endif


/* ---------------------------------------------------------------------------
 * disable warning C4100: unreferenced formal parameter
 */
#if defined(_MSC_VER) || defined(__INTEL_COMPILER)
#define UNUSED_PARAMETER(P)     (P)   /* same as UNREFERENCED_PARAMETER */
#else
#define UNUSED_PARAMETER(P)
#endif


/**
 * ===========================================================================
 * const defines
 * ===========================================================================
 */
/* ---------------------------------------------------------------------------
 * Specifies the number of bits per pixel that XRDENC uses. This is also the
 * bit depth that XRDENC encodes in. If this value is > 8, XRDENC will read
 * two bytes of input data for each pixel sample, and expect the upper
 * (16-XRD_BIT_DEPTH) bits to be zero.
 * Note: The flag XRD_CSP_HIGH_DEPTH must be used to specify the
 * colorspace depth as well.
 */
#define WORD_SIZE               sizeof(void*)

/**
 * ===========================================================================
 * const defines
 * ===========================================================================
 */

#if defined(__GNUC__) && (__GNUC__ > 3 || __GNUC__ == 3 && __GNUC_MINOR__ > 0)
#define UNINIT(x)               x = x
#define UNUSED                  __attribute__((unused))
#define ALWAYS_INLINE           __attribute__((always_inline)) inline
#define NOINLINE                __attribute__((noinline))
#define MAY_ALIAS               __attribute__((may_alias))
#define xrd_constant_p(x)   __builtin_constant_p(x)
#define xrd_nonconstant_p(x)    (!__builtin_constant_p(x))
#define INLINE                  __inline
#else
#define UNINIT(x)               x
#if defined(__ICL)
#define ALWAYS_INLINE           __forceinline
#define NOINLINE                __declspec(noinline)
#else
#define ALWAYS_INLINE           INLINE
#define NOINLINE
#endif
#define UNUSED
#define MAY_ALIAS
#define xrd_constant_p(x)       0
#define xrd_nonconstant_p(x)    0
#endif

#if defined(__ICL) || defined(_MSC_VER)
#define INLINE                  __inline
#define strcasecmp              _stricmp
#define strncasecmp             _strnicmp
#if !HAVE_POSIXTHREAD
#define strtok_r                strtok_s
#endif
#define S_ISREG(x)              (((x) & S_IFMT) == S_IFREG)
#endif

#if defined(_MSC_VER) || defined(__INTEL_COMPILER)
#undef  fseek
#undef  ftell
#undef  isfinite
#undef  fileno
#define fseek                   _fseeki64
#define ftell                   _ftelli64
#define isfinite                _finite
#define fileno                  _fileno
#endif

#if !defined(isfinite) && (SYS_OPENBSD || SYS_SunOS)
#define isfinite                finite
#endif


/* ---------------------------------------------------------------------------
 * align
 */
/* align a pointer */
#  define CACHE_LINE_SIZE       32    /* for x86-64 and x86 */
#  define ALIGN_POINTER(p)      (p) = (uint8_t *)((intptr_t)((p) + (CACHE_LINE_SIZE - 1)) & (~(intptr_t)(CACHE_LINE_SIZE - 1)))
#  define CACHE_LINE_256B       32    /* for x86-64 and x86 */
#  define ALIGN_256_PTR(p)      (p) = (uint8_t *)((intptr_t)((p) + (CACHE_LINE_256B - 1)) & (~(intptr_t)(CACHE_LINE_256B - 1)))

#if defined(_MSC_VER)
#pragma warning(disable:4324)   /* disable warning C4324:  __declspec(align())ṹ */
#define DECLARE_ALIGNED(var, n) __declspec(align(n)) var
#elif defined(__INTEL_COMPILER)
#define DECLARE_ALIGNED(var, n) var __declspec(align(n))
#else
#define DECLARE_ALIGNED(var, n) var __attribute__((aligned(n)))
#endif
#define ALIGN32(var)            DECLARE_ALIGNED(var, 32)
#define ALIGN16(var)            DECLARE_ALIGNED(var, 16)
#define ALIGN8(var)             DECLARE_ALIGNED(var, 8)


// ARM compiliers don't reliably align stack variables
// - EABI requires only 8 byte stack alignment to be maintained
// - gcc can't align stack variables to more even if the stack were to be correctly aligned outside the function
// - armcc can't either, but is nice enough to actually tell you so
// - Apple gcc only maintains 4 byte alignment
// - llvm can align the stack, but only in svn and (unrelated) it exposes bugs in all released GNU binutils...

#define ALIGNED_ARRAY_EMU( mask, type, name, sub1, ... )\
    uint8_t name##_u [sizeof(type sub1 __VA_ARGS__) + mask]; \
    type (*name) __VA_ARGS__ = (void*)((intptr_t)(name##_u+mask) & ~mask)

#if ARCH_ARM && SYS_MACOSX
#define ALIGNED_ARRAY_8( ... ) ALIGNED_ARRAY_EMU( 7, __VA_ARGS__ )
#else
#define ALIGNED_ARRAY_8( type, name, sub1, ... ) \
    ALIGN8( type name sub1 __VA_ARGS__ )
#endif

#if ARCH_ARM
#define ALIGNED_ARRAY_16( ... ) ALIGNED_ARRAY_EMU( 15, __VA_ARGS__ )
#else
#define ALIGNED_ARRAY_16( type, name, sub1, ... ) \
    ALIGN16( type name sub1 __VA_ARGS__ )
#endif

#define EXPAND(x)               x

#if defined(STACK_ALIGNMENT) && STACK_ALIGNMENT >= 32
#define ALIGNED_ARRAY_32( type, name, sub1, ... ) \
    ALIGN32( type name sub1 __VA_ARGS__ )
#else
#define ALIGNED_ARRAY_32(...)   EXPAND(ALIGNED_ARRAY_EMU(31, __VA_ARGS__))
#endif

#define ALIGNED_ARRAY_64(...)   EXPAND(ALIGNED_ARRAY_EMU(63, __VA_ARGS__))

/* For AVX2 */
#if ARCH_X86 || ARCH_X86_64
#define NATIVE_ALIGN            32
#define ALIGNED_N               ALIGN32
#define ALIGNED_ARRAY_N         ALIGNED_ARRAY_32
#else
#define NATIVE_ALIGN            16
#define ALIGNED_N               ALIGN16
#define ALIGNED_ARRAY_N         ALIGNED_ARRAY_16
#endif


/* ---------------------------------------------------------------------------
 * threads
 */
#if HAVE_BEOSTHREAD
#include <kernel/OS.h>
#define xrd_pthread_t       thread_id
static int ALWAYS_INLINE
xrd_pthread_create(xrd_pthread_t *t, void *a, void *(*f)(void *), void *d)
{
    *t = spawn_thread(f, "", 10, d);
    if (*t < B_NO_ERROR) {
        return -1;
    }
    resume_thread(*t);
    return 0;
}

#define xrd_pthread_join(t,s) \
{\
    long tmp; \
    wait_for_thread(t,(s)?(long*)(*(s)):&tmp);\
}

#elif HAVE_POSIXTHREAD
#if defined(_MSC_VER) || defined(__ICL)
#if _MSC_VER >= 1900
#define HAVE_STRUCT_TIMESPEC    1       /* for struct timespec */
#endif
#pragma comment(lib, "pthread_lib.lib")
#endif
#include <pthread.h>
#define xrd_pthread_t                   pthread_t
#define xrd_pthread_create              pthread_create
#define xrd_pthread_join                pthread_join
#define xrd_pthread_mutex_t             pthread_mutex_t
#define xrd_pthread_mutex_init          pthread_mutex_init
#define xrd_pthread_mutex_destroy       pthread_mutex_destroy
#define xrd_pthread_mutex_lock          pthread_mutex_lock
#define xrd_pthread_mutex_unlock        pthread_mutex_unlock
#define xrd_pthread_cond_t              pthread_cond_t
#define xrd_pthread_cond_init           pthread_cond_init
#define xrd_pthread_cond_destroy        pthread_cond_destroy
#define xrd_pthread_cond_signal         pthread_cond_signal
#define xrd_pthread_cond_broadcast      pthread_cond_broadcast
#define xrd_pthread_cond_wait           pthread_cond_wait
#define xrd_pthread_attr_t              pthread_attr_t
#define xrd_pthread_attr_init           pthread_attr_init
#define xrd_pthread_attr_destroy        pthread_attr_destroy
#define xrd_pthread_attr_setdetachstate pthread_attr_setdetachstate
#define xrd_pthread_num_processors_np   pthread_num_processors_np
#define XRD_PTHREAD_MUTEX_INITIALIZER   PTHREAD_MUTEX_INITIALIZER

#elif HAVE_WIN32THREAD
#include "win32thread.h"

#else
#define xrd_pthread_t                   int
#define xrd_pthread_create(t,u,f,d)     0
#define xrd_pthread_join(t,s)
#endif // HAVE_*THREAD

#if !HAVE_POSIXTHREAD && !HAVE_WIN32THREAD
#define xrd_pthread_mutex_t             int
#define xrd_pthread_mutex_init(m,f)     0
#define xrd_pthread_mutex_destroy(m)
#define xrd_pthread_mutex_lock(m)
#define xrd_pthread_mutex_unlock(m)
#define xrd_pthread_cond_t              int
#define xrd_pthread_cond_init(c,f)      0
#define xrd_pthread_cond_destroy(c)
#define xrd_pthread_cond_broadcast(c)
#define xrd_pthread_cond_wait(c,m)
#define xrd_pthread_attr_t              int
#define xrd_pthread_attr_init(a)        0
#define xrd_pthread_attr_destroy(a)
#define XRD_PTHREAD_MUTEX_INITIALIZER   0
#endif

#if HAVE_POSIXTHREAD
#if SYS_WINDOWS
#define xrd_lower_thread_priority(p) \
{\
    xrd_pthread_t handle = pthread_self();\
    struct sched_param sp;\
    int policy = SCHED_OTHER;\
    pthread_getschedparam(handle, &policy, &sp);\
    sp.sched_priority -= p;\
    pthread_setschedparam(handle, policy, &sp);\
}

#else
#include <unistd.h>
#define xrd_lower_thread_priority(p) { UNUSED int nice_ret = nice(p); }
#endif /* SYS_WINDOWS */
#elif HAVE_WIN32THREAD
#define xrd_lower_thread_priority(p) SetThreadPriority(GetCurrentThread(), XRD_MAX(-2, -p))
#else
#define xrd_lower_thread_priority(p)
#endif


/**
 * ===========================================================================
 * type defines
 * ===========================================================================
 */
/* custom spin lock - the initial value of spin lock must be 0 */
#if (defined(__ICL) || defined(_MSC_VER)) && defined(_WIN32)
typedef void                   *semaphore_t;
typedef volatile long           spinlock_t;
#  define SPIN_INIT(lock)       (lock) = 0
#  define SPIN_LOCK(lock)       while (InterlockedExchange(&(lock), 1) == 1) Sleep(0)
#  define SPIN_UNLOCK(lock)     InterlockedExchange(&(lock), 0)
#  define SPIN_DESTROY(lock)    
#elif defined(__ARM_ARCH_7A__) || defined(__ARM_ARCH_7__)
#include <unistd.h>
#include <semaphore.h>
typedef sem_t                   semaphore_t;
typedef pthread_mutex_t         spinlock_t;
#  define Sleep(x)              usleep(x * 1000)
#  define SPIN_INIT(lock)       pthread_mutex_init(&(lock), NULL)
#  define SPIN_LOCK(lock)       pthread_mutex_lock(&(lock))
#  define SPIN_UNLOCK(lock)     pthread_mutex_unlock(&(lock))
#  define SPIN_DESTROY(lock)    pthread_mutex_destroy(&(lock))
#else
#include <unistd.h>
#include <semaphore.h>
#include <pthread.h>
typedef sem_t                   semaphore_t;
typedef pthread_spinlock_t      spinlock_t;
#  define Sleep(x)              usleep(x * 1000)
#  define SPIN_INIT(lock)       pthread_spin_init(&(lock), PTHREAD_PROCESS_PRIVATE)
#  define SPIN_LOCK(lock)       pthread_spin_lock(&(lock))
#  define SPIN_UNLOCK(lock)     pthread_spin_unlock(&(lock))
#  define SPIN_DESTROY(lock)    pthread_spin_destroy(&(lock))
#endif


/**
 * ===========================================================================
 * inline functions
 * ===========================================================================
 */

#if defined(__GNUC__) && (__GNUC__ > 3 || __GNUC__ == 3 && __GNUC_MINOR__ > 3)
#define xrd_clz(x)      __builtin_clz(x)
#define xrd_ctz(x)      __builtin_ctz(x)
#elif defined(_MSC_VER) && defined(_WIN32)
static int ALWAYS_INLINE xrd_clz(const uint32_t x)
{
    DWORD r;
    _BitScanReverse(&r, (DWORD)x);
    return (r ^ 31);
}

static int ALWAYS_INLINE xrd_ctz(const uint32_t x)
{
    DWORD r;
    _BitScanForward(&r, (DWORD)x);
    return r;
}

#else
static int ALWAYS_INLINE xrd_clz(uint32_t x)
{
    static uint8_t lut[16] = {4, 3, 2, 2, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0};
    int y, z = (((x >> 16) - 1) >> 27) & 16;
    x >>= z ^ 16;
    z += y = ((x - 0x100) >> 28) & 8;
    x >>= y ^ 8;
    z += y = ((x - 0x10) >> 29) & 4;
    x >>= y ^ 4;
    return z + lut[x];
}

static int ALWAYS_INLINE xrd_ctz(uint32_t x)
{
    static uint8_t lut[16] = {4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0};
    int y, z = (((x & 0xffff) - 1) >> 27) & 16;
    x >>= z;
    z += y = (((x & 0xff) - 1) >> 28) & 8;
    x >>= y;
    z += y = (((x & 0xf) - 1) >> 29) & 4;
    x >>= y;
    return z + lut[x & 0xf];
}
#endif


/* ---------------------------------------------------------------------------
 * prefetch
 */
#if defined(__GNUC__) && (__GNUC__ > 3 || __GNUC__ == 3 && __GNUC_MINOR__ > 1) &&\
      (ARCH_X86 || ARCH_X86_64 || ARCH_ARM || ARCH_PPC)
#define xrd_prefetch(x)     __builtin_prefetch(x)
#elif defined(_MSC_VER)
#define xrd_prefetch(x)     _mm_prefetch((const char*)(x), _MM_HINT_T0)
#else
#define xrd_prefetch(x)
#endif


/* ---------------------------------------------------------------------------
 * log2/log2f
 */
#if !HAVE_LOG2F
#define log2f(x)    (logf(x)/0.693147180559945f)
#define log2(x)     (log(x)/0.693147180559945)
#endif


/* ---------------------------------------------------------------------------
 * file
 */
static int ALWAYS_INLINE xrd_is_regular_file(int filehandle)
{
    struct stat file_stat;
    if (fstat(filehandle, &file_stat)) {
        return -1;
    }
    return S_ISREG(file_stat.st_mode);
}

static int ALWAYS_INLINE xrd_is_regular_file_ex(FILE *filehandle)
{
    struct stat file_stat;
    if (fstat(fileno(filehandle), &file_stat)) {
        return -1;
    }
    return S_ISREG(file_stat.st_mode);
}

static int ALWAYS_INLINE xrd_is_regular_file_path(const char *filename)
{
    struct stat file_stat;
    if (stat(filename, &file_stat)) {
        return -1;
    }
    return S_ISREG(file_stat.st_mode);
}

#endif /* __XRD_OSDEP_H__ */
