/*
 * yuv_writer.c
 *
 * Description of this file:
 *    YUV Writing functions definition of the xRD library
 *
 * --------------------------------------------------------------------------
 *
 *    xRD - video encoder of AVS2/IEEE1857.4 video coding standard
 *    Copyright (C) 2018~ VCL, NELVT, Peking University
 *
 *    Authors: Falei LUO <falei.luo@gmail.com>
 *             etc.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111, USA.
 *
 *    This program is also available under a commercial proprietary license.
 *    For more information, contact us at sswang @ pku.edu.cn.
 */

#include "common.h"
#include "wrapper.h"
#include "encoder.h"

/* ---------------------------------------------------------------------------
 */
static INLINE
void dump_yuv_out_10bit(FILE *fp, xrd_frame_t *frame, int img_w, int img_h)
{
    int j;

    if (fp != NULL) {
        for (j = 0; j < img_h; j++) {
            fwrite(frame->planes[0] + j * frame->i_stride[0], img_w << 1, 1, fp);
        }

        if (frame->i_plane == 3) {
            for (j = 0; j < (img_h >> 1); j++) {
                fwrite(frame->planes[1] + j * frame->i_stride[1], img_w, 1, fp);
            }

            for (j = 0; j < (img_h >> 1); j++) {
                fwrite(frame->planes[2] + j * frame->i_stride[2], img_w, 1, fp);
            }
        }
    }
}


/* ---------------------------------------------------------------------------
 */
static INLINE
void dump_yuv_out(xrd_t *h, FILE *fp, xrd_frame_t *frame, int img_w, int img_h)
{
    int j;

    if (fp != NULL) {
        static uint8_t p_buffer[8192];
        if (h->param.input_sample_bit_depth == 8) {
            const int shift_out  = h->param.sample_bit_depth - h->param.input_sample_bit_depth;
            const int offset_out = 1 << (shift_out - 1);
            int i;

            for (j = 0; j < img_h; j++) {
                for (i = 0; i < img_w; i++) {
                    p_buffer[i] = (uint8_t)((frame->planes[0][j * frame->i_stride[0] + i] + offset_out) >> shift_out);
                }
                fwrite(p_buffer, img_w, 1, fp);
            }

            if (frame->i_plane == 3) {
                for (j = 0; j < (img_h >> 1); j++) {
                    for (i = 0; i < (img_w >> 1); i++) {
                        p_buffer[i] = (uint8_t)((frame->planes[1][j * frame->i_stride[1] + i] + offset_out) >> shift_out);
                    }
                    fwrite(p_buffer, img_w >> 1, 1, fp);
                }

                for (j = 0; j < (img_h >> 1); j++) {
                    for (i = 0; i < (img_w >> 1); i++) {
                        p_buffer[i] = (uint8_t)((frame->planes[2][j * frame->i_stride[2] + i] + offset_out) >> shift_out);
                    }
                    fwrite(p_buffer, img_w >> 1, 1, fp);
                }
            }
        } else {
            dump_yuv_out_10bit(fp, frame, img_w, img_h);
        }
    }
}

/* ---------------------------------------------------------------------------
 */
void encoder_write_rec_frame(xrd_handler_t *h_mgr)
{
    xrd_t        *h     = h_mgr->p_coder;
    xrd_frame_t **DPB   = h_mgr->dpb.frames;
    int size_dpb = h_mgr->dpb.num_frames;
    int i = 0;
    int j;

    xrd_pthread_mutex_lock(&h_mgr->mutex);   /* lock */

    while (i < size_dpb) {
        int next_output_frame_idx;
        xrd_frame_t  *frame = DPB[i];
        if (frame == NULL) {
            i++;
            continue;
        }

        xrd_pthread_mutex_lock(&frame->mutex);  /* lock */

        next_output_frame_idx = get_next_frame_id(h_mgr->i_output);
        if (frame->i_frame == next_output_frame_idx) {
            /* has the frame already been reconstructed ? */
            for (j = 0; j < h->i_height_in_lcu; j++) {
                if (frame->num_lcu_coded_in_row[j] < h->i_width_in_lcu) {
                    break;
                }
            }

            if (j < h->i_height_in_lcu) {
                /* frame doesn't finish reconstruction */
                xrd_pthread_mutex_unlock(&frame->mutex);   /* unlock */
                break;
            }

            /* update output frame index */
            h_mgr->i_output = next_output_frame_idx;

#if XRD_DUMP_REC
            dump_yuv_out(h, h_mgr->h_rec_file, frame, h->param.org_width, h->param.org_height);
#endif //if XRD_DUMP_REC
            xrd_pthread_mutex_unlock(&frame->mutex);   /* unlock */

            /* start over for the next reconstruction frame */
            i = 0;
            continue;
        }

        xrd_pthread_mutex_unlock(&frame->mutex);    /* unlock */
        i++;
    }

    xrd_pthread_mutex_unlock(&h_mgr->mutex);     /* unlock */
}
